import copy
import datetime
import itertools
import logging
import os
from typing import Any

from eixgnn.eixgnn import EiXGNN
from scgnn.scgnn import SCGNN
from torch_geometric import seed_everything
from torch_geometric.data import Batch, Data
from torch_geometric.data.makedirs import makedirs
from torch_geometric.explain import Explainer
from torch_geometric.explain.config import ThresholdConfig
from torch_geometric.explain.explanation import Explanation
from torch_geometric.graphgym.config import cfg
from torch_geometric.graphgym.loader import create_dataset
from torch_geometric.graphgym.model_builder import cfg, create_model
from torch_geometric.graphgym.utils.device import auto_select_device
from torch_geometric.loader.dataloader import DataLoader
from yacs.config import CfgNode as CN

from explaining_framework.config.explainer_config.eixgnn_config import \
    eixgnn_cfg
from explaining_framework.config.explainer_config.scgnn_config import scgnn_cfg
from explaining_framework.config.explaining_config import explaining_cfg
from explaining_framework.explainers.wrappers.from_captum import CaptumWrapper
from explaining_framework.explainers.wrappers.from_graphxai import \
    GraphXAIWrapper
from explaining_framework.explainers.wrappers.from_pyg import PYGWrapper
from explaining_framework.metric.accuracy import Accuracy
from explaining_framework.metric.base import Metric
from explaining_framework.metric.fidelity import Fidelity
from explaining_framework.metric.robust import Attack
from explaining_framework.metric.sparsity import Sparsity
from explaining_framework.stats.graph.graph_stat import GraphStat
from explaining_framework.utils.explaining.load_ckpt import (LoadModelInfo,
                                                             _load_ckpt)
from explaining_framework.utils.explanation.adjust import Adjust
from explaining_framework.utils.explanation.io import (
    _get_explanation, _load_explanation, _save_explanation,
    explanation_verification, get_pred, is_empty_graph)
from explaining_framework.utils.io import (dump_cfg, is_exists,
                                           obj_config_to_log,
                                           obj_config_to_str, read_json,
                                           set_printing, write_json,
                                           write_yaml)

all__captum = [
    "LRP",
    "DeepLift",
    "DeepLiftShap",
    "FeatureAblation",
    "FeaturePermutation",
    "GradientShap",
    "GuidedBackprop",
    "GuidedGradCam",
    "InputXGradient",
    # "IntegratedGradients",
    "Lime",
    "Occlusion",
    "Saliency",
]

all__graphxai = [
    "CAM",
    "GradCAM",
    "GNN_LRP",
    "GradExplainer",
    "GuidedBackPropagation",
    "IntegratedGradients",
    "PGExplainer",
    "PGMExplainer",
    "RandomExplainer",
    "SubgraphX",
    "GraphMASK",
    "GNNExplainer",
]
all__pyg = [
    # "PGExplainer",
    # "GNNExplainer",
]

all__own = ["EIXGNN", "SCGNN"]

all_fidelity = [
    "fidelity_plus",
    "fidelity_minus",
    "fidelity_plus_prob",
    "fidelity_minus_prob",
    "infidelity_KL",
    "characterization",
    "characterization_prob",
]
all_accuracy = [
    "precision_score",
    "jaccard_score",
    "roc_auc_score",
    "f1_score",
    "accuracy_score",
]

all_robust = [
    "gaussian_noise",
    "add_edge",
    "remove_edge",
    "remove_node",
    "pgd",
    "fgsm",
    "no_attack",
]
all_sparsity = ["l0"]

adjust_pattern = "ranps"
all_adjusts_filters = [
    "".join(filters)
    for i in range(len(adjust_pattern) + 1)
    for filters in itertools.permutations(adjust_pattern, i)
]

all_threshold_type = ["topk_hard", "hard", "topk"]


class ExplainingOutline(object):
    def __init__(self, explaining_cfg_path: str):
        self.explaining_cfg_path = explaining_cfg_path
        self.explaining_cfg = None
        self.explainer_cfg_path = None
        self.explainer_cfg = None
        self.explaining_algorithm = None
        self.cfg = None
        self.model = None
        self.dataset = None
        self.model_info = None
        self.metrics = None
        self.attacks = None
        self.model_signature = None
        self.indexes = None
        self.sparsities = None
        self.fidelities = None
        self.accuracies = None
        self.explaining_algorithm = None
        self.explainer = None
        self.adjusts = None
        self.thresholds_configs = None
        self.graphstat = None

        self.load_explaining_cfg()
        self.load_model_info()
        self.load_cfg()
        self.load_cfg_to_explaining_cfg()
        self.load_dataset()
        self.load_model()
        self.load_model_to_hardware()
        self.load_explainer_cfg()
        self.load_explaining_algorithm()
        self.load_explainer()
        self.load_metric()
        self.load_attack()
        self.load_dataset_to_dataloader()
        self.load_indexes()
        self.load_adjust()
        self.load_threshold()
        self.load_graphstat()
        self.setup_experiment()

        seed_everything(self.explaining_cfg.seed)

    def load_model_to_hardware(self):
        auto_select_device()
        device = self.cfg.accelerator
        self.model = self.model.to(device)

    def get_data(self):
        if self.dataset is None:
            self.load_dataset()
        try:
            item = next(self.dataset)
            device = self.cfg.accelerator
            item = item.to(device)
            return item
        except StopIteration:
            return None

    def load_indexes(self):

        item = self.explaining_cfg.dataset.item
        if isinstance(item, (list, int)):
            indexes = item
        else:
            indexes = list(range(len(self.dataset)))
        self.indexes = iter(indexes)

    def get_index(self):
        if self.indexes is None:
            self.load_indexes()
        try:
            item = next(self.indexes)
            return item
        except StopIteration:
            return None

    def get_item(self):
        item = self.get_data()
        index = self.get_index()
        return item, index

    def load_model_info(self):
        info = LoadModelInfo(
            dataset_name=self.explaining_cfg.dataset.name,
            model_dir=self.explaining_cfg.model.path,
            which=self.explaining_cfg.model.ckpt,
        )
        if info.list_xp() is None:
            raise ValueError

        self.model_info = info.set_info()
        self.model_signature = info.get_model_signature()

    def load_cfg(self):
        cfg.merge_from_file(self.model_info["cfg_path"])
        self.cfg = cfg

    def load_explaining_cfg(self):
        explaining_cfg.merge_from_file(self.explaining_cfg_path)
        self.explaining_cfg = explaining_cfg

    def load_cfg_to_explaining_cfg(self):
        if self.cfg is None:
            self.load_cfg()
        if self.explaining_cfg is None:
            self.load_explaining_cfg()
        self.explaining_cfg.model_config.task_level = self.cfg.dataset.task

    def load_explainer_cfg(self):
        if self.explaining_cfg is None:
            self.load_explaining_cfg()
        else:
            if self.explaining_cfg.explainer.cfg == "default":
                if self.explaining_cfg.explainer.name == "EIXGNN":
                    self.explainer_cfg = copy.copy(eixgnn_cfg)
                elif self.explaining_cfg.explainer.name == "SCGNN":
                    self.explainer_cfg = copy.copy(scgnn_cfg)
                else:
                    self.explainer_cfg = CN()
            else:
                if self.explaining_cfg.explainer.name == "EIXGNN":
                    eixgnn_cfg.merge_from_file(self.explaining_cfg.explainer.cfg)
                    self.explainer_cfg = eixgnn_cfg
                elif self.explaining_cfg.explainer.name == "SCGNN":
                    scgnn_cfg.merge_from_file(self.explaining_cfg.explainer.cfg)
                    self.explainer_cfg = scgnn_cfg

    def load_model(self):
        if self.cfg is None:
            self.load_cfg()
        auto_select_device()
        self.model = create_model()
        self.model = _load_ckpt(self.model, self.model_info["ckpt_path"])
        if self.model is None:
            raise ValueError("Model ckpt has not been loaded, ckpt file not found")
        self.model = self.model.eval()
        self.model.explain = True

    def load_dataset(self):
        if self.cfg is None:
            self.load_cfg()
        if self.explaining_cfg is None:
            self.load_explaining_cfg()
        if self.explaining_cfg.dataset.name != self.cfg.dataset.name:
            raise ValueError(
                f"Expecting that the dataset to perform explanation on is the same as the model has trained on. Get {self.explaining_cfg.dataset.name} for explanation part, and {self.cfg.dataset.name} for the model."
            )
        self.dataset = create_dataset()
        item = self.explaining_cfg.dataset.item
        if isinstance(item, int):
            self.dataset = self.dataset[item : item + 1]
        elif isinstance(item, list):
            self.dataset = self.dataset[item]

    def load_dataset_to_dataloader(self, to_iter=True):
        self.dataset = DataLoader(dataset=self.dataset, shuffle=False, batch_size=1)
        if to_iter:
            self.dataset = iter(self.dataset)

    def reload_dataset(self):
        self.load_dataset()
        self.load_indexes()

    def reload_dataloader(self):
        self.load_dataset()
        self.load_dataset_to_dataloader()
        self.load_indexes()

    def load_explaining_algorithm(self):
        self.load_explainer_cfg()
        if self.model is None:
            self.load_model()
        if self.dataset is None:
            self.load_dataset()

        name = self.explaining_cfg.explainer.name
        if name in all__graphxai:
            explaining_algorithm = GraphXAIWrapper(
                name,
                in_channels=self.dataset.num_classes,
                criterion=self.cfg.model.loss_fun,
            )
        elif name in all__captum:
            explaining_algorithm = CaptumWrapper(name)
        elif name in all__pyg:
            explaining_algorithm = PYGWrapper(name)
        elif name in all__own:
            if name == "EIXGNN":
                explaining_algorithm = EiXGNN(
                    L=self.explainer_cfg.L,
                    p=self.explainer_cfg.p,
                    importance_sampling_strategy=self.explainer_cfg.importance_sampling_strategy,
                    domain_similarity=self.explainer_cfg.domain_similarity,
                    signal_similarity=self.explainer_cfg.signal_similarity,
                    shap_val_approx=self.explainer_cfg.shapley_value_approx,
                )
            elif name == "SCGNN":
                explaining_algorithm = SCGNN(
                    depth=self.explainer_cfg.depth,
                    interest_map_norm=self.explainer_cfg.interest_map_norm,
                    score_map_norm=self.explainer_cfg.score_map_norm,
                    target_baseline=self.explainer_cfg.target_baseline,
                )
        elif name is None:
            explaining_algorithm = None
        else:
            raise ValueError(f"{name_} Metric is not supported yet")
        self.explaining_algorithm = explaining_algorithm

    def load_explainer(self):
        if self.explaining_algorithm is None:
            self.load_explaining_algorithm()
        explainer = Explainer(
            model=self.model,
            algorithm=self.explaining_algorithm,
            explainer_config=dict(
                explanation_type=self.explaining_cfg.explanation_type,
                node_mask_type="object",
                edge_mask_type="object",
            ),
            model_config=dict(
                mode="regression",
                task_level=self.explaining_cfg.model_config.task_level,
                return_type=self.explaining_cfg.model_config.return_type,
            ),
        )
        self.explainer = explainer

    def load_fidelity(self):
        if self.cfg is None:
            self.load_cfg()
        if self.explaining_cfg is None:
            self.load_explaining_cfg()
        name = self.explaining_cfg.metrics.fidelity.name
        if name == "all":
            all_metrics = [
                Fidelity(name=name, model=self.model) for name in all_fidelity
            ]
        elif isinstance(name, str):
            if name in all_fidelity:
                all_metrics = [Fidelity(name=name, model=self.model)]
            else:
                raise ValueError(
                    f"This fidelity metric {name} is nor supported yet. Supported are {all_fidelity}"
                )
        elif isinstance(name, list):
            all_metrics = [
                Fidelity(name=name, model=self.model)
                for name_ in name
                if name_ in all_fidelity
            ]
        elif name is None:
            all_metrics = []
        self.fidelities = all_metrics

    def load_sparsity(self):
        if self.cfg is None:
            self.load_cfg()
        if self.explaining_cfg is None:
            self.load_explaining_cfg()
        name = self.explaining_cfg.metrics.sparsity.name
        if name == "all":
            all_metrics = [Sparsity(name=name) for name in all_sparsity]
        elif isinstance(name, str):
            if name in all_sparsity:
                all_metrics = [Sparsity(name=name)]
            else:
                raise ValueError(
                    f"This sparsity metric {name} is nor supported yet. Supported are {all_sparsity}"
                )
        elif isinstance(name, list):
            all_metrics = [
                Sparsity(name=name) for name_ in name if name_ in all_sparsity
            ]
        elif name is None:
            all_metrics = []
        self.sparsities = all_metrics

    def load_accuracy(self):
        if self.cfg is None:
            self.load_cfg()
        if self.explaining_cfg is None:
            self.load_explaining_cfg()

        if self.explaining_cfg.dataset.name == "BASHAPES":
            name = self.explaining_cfg.metrics.accuracy.name
            if name == "all":
                all_metrics = [Accuracy(name=name) for name in all_accuracy]
            elif isinstance(name, str):
                if name in all_accuracy:
                    all_metrics = [Accuracy(name=name)]
                else:
                    raise ValueError(
                        f"This accuracy metric {name} is nor supported yet. Supported are {all_accuracy}"
                    )
            elif isinstance(name, list):
                all_metrics = [
                    Accuracy(name=name) for name_ in name if name_ in all_accuracy
                ]
            elif name is None:
                all_metrics = []
            else:
                raise ValueError(
                    f"Provided dataset needs explanation groundtruths for using Accuracies metric, e.g BASHAPES dataset"
                )
        else:
            all_metrics = []
        self.accuracies = all_metrics

    def load_metric(self):
        if self.cfg is None:
            self.load_cfg()
        if self.explaining_cfg is None:
            self.load_explaining_cfg()
        if self.accuracies is None:
            self.load_accuracy()
        if self.sparsities is None:
            self.load_sparsity()
        if self.fidelities is None:
            self.load_fidelity()

        self.metrics = self.fidelities + self.accuracies + self.sparsities

    def load_attack(self):
        if self.cfg is None:
            self.load_cfg()
        if self.explaining_cfg is None:
            self.load_explaining_cfg()
        name = self.explaining_cfg.attack.name
        if name == "all":
            all_metrics = [Attack(name=name, model=self.model) for name in all_robust]
        elif isinstance(name, str):
            if name in all_robust:
                all_metrics = [Attack(name=name, model=self.model)]
            else:
                raise ValueError(
                    f"This Attack metric {name} is not supported yet. Supported are {all_robust}"
                )
        elif isinstance(name, list):
            all_metrics = [
                Attack(name=name, model=self.model)
                for name_ in name
                if name_ in all_robust
            ]
        elif name is None:
            all_metrics = []
        self.attacks = all_metrics

    def load_adjust(self):
        if self.explaining_cfg is None:
            self.load_explaining_cfg()
        strategy = self.explaining_cfg.adjust.strategy
        if strategy == "all":
            self.adjusts = [Adjust(strategy=strat) for strat in all_adjusts_filters]
        elif isinstance(strategy, str):
            if strategy in all_adjusts_filters:
                all_metrics = [Adjust(strategy=strategy)]
            else:
                raise ValueError(
                    f"This Adjust metric {name} is not supported yet. Supported are {all_adjusts_filters}"
                )
        elif isinstance(strategy, list):
            all_metrics = [
                Adjust(strategy=name_)
                for name_ in strategy
                if name_ in all_adjusts_filters
            ]
        elif name is None:
            all_metrics = []
        self.adjusts = all_metrics

    def load_threshold(self):
        if self.explaining_cfg is None:
            self.load_explaining_cfg()
        threshold_type = self.explaining_cfg.threshold.config.type
        if threshold_type == "all":
            th_hard = [
                {"threshold_type": "hard", "value": th_value}
                for th_value in self.explaining_cfg.threshold.value.hard
            ]
            th_topk = [
                {"threshold_type": th_type, "value": th_value}
                for th_value in self.explaining_cfg.threshold.value.topk
                for th_type in all_threshold_type
                if "topk" in th_type
            ]
            all_threshold = th_hard + th_topk
        elif isinstance(threshold_type, str):
            if threshold_type in all_threshold_type:
                if "topk" in threshold_type:
                    all_threshold = [
                        {
                            "threshold_type": threshold_type,
                            "value": threshold_value,
                        }
                        for threshold_value in self.explaining_cfg.threshold.value.topk
                    ]
                elif threshold_type == "hard":
                    all_threshold = [
                        {
                            "threshold_type": threshold_type,
                            "value": threshold_value,
                        }
                        for threshold_value in self.explaining_cfg.threshold.value.hard
                    ]
        elif isinstance(threshold_type, list):
            all_threshold = []
            for tf_type in threshold_type:
                if "topk" in th_type:
                    all_threshold.expend(
                        [
                            {
                                "threshold_type": threshold_type,
                                "value": threshold_value,
                            }
                            for threshold_value in self.explaining_cfg.threshold.value.topk
                        ]
                    )
                elif th_type == "hard":
                    all_threshold.expend(
                        [
                            {
                                "threshold_type": threshold_type,
                                "value": threshold_value,
                            }
                            for threshold_value in self.explaining_cfg.threshold.value.hard
                        ]
                    )

        elif threshold_type is None:
            all_threshold = []
        self.thresholds_configs = all_threshold

    def set_explainer_threshold_config(self, threshold_config):
        self.explainer.threshold_config = ThresholdConfig.cast(threshold_config)

    def load_graphstat(self):
        self.graphstat = GraphStat()

    def get_explanation(self, item: Data, path: str):
        if is_exists(path):
            if self.explaining_cfg.explainer.force:
                explanation = _get_explanation(self.explainer, item)
                if explanation is None:
                    logging.warning(
                        " EXP::Generated; Path %s; FAILED",
                        (path),
                    )
                else:
                    logging.debug(
                        "EXP::Generated; Path %s; SUCCEEDED",
                        (path),
                    )
            else:
                explanation = _load_explanation(path)
                logging.debug(
                    "EXP::Loaded; Path %s; SUCCEEDED",
                    (path),
                )
                explanation = explanation.to(self.cfg.accelerator)
        else:
            explanation = _get_explanation(self.explainer, item)
            get_pred(self.explainer, explanation)
            _save_explanation(explanation, path)
            logging.debug(
                "EXP::Generated; Path %s; SUCCEEDED",
                (path),
            )

        return explanation

    def get_adjust(self, adjust: Adjust, item: Explanation, path: str):
        if is_exists(path):
            if self.explaining_cfg.explainer.force:
                exp_adjust = adjust.forward(item)
                logging.debug(
                    "ADJUST::Generated; Path %s; SUCCEEDED",
                    (path),
                )

            else:
                exp_adjust = _load_explanation(path)
                logging.debug(
                    "ADJUST::Loaded; Path %s; SUCCEEDED",
                    (path),
                )

        else:
            exp_adjust = adjust.forward(item)
            get_pred(self.explainer, exp_adjust)
            _save_explanation(exp_adjust, path)
            logging.debug(
                "ADJUST::Generated; Path %s; SUCCEEDED",
                (path),
            )
        return exp_adjust

    def get_threshold(self, item: Explanation, path: str):
        if is_exists(path):
            if self.explaining_cfg.explainer.force:
                exp_threshold = self.explainer._post_process(item)
                logging.debug(
                    "THRESHOLD::Generated; Path %s; SUCCEEDED",
                    (path),
                )
            else:
                exp_threshold = _load_explanation(path)
                logging.debug(
                    "THRESHOLD::Loaded; Path %s; SUCCEEDED",
                    (path),
                )
        else:
            exp_threshold = self.explainer._post_process(item)
            get_pred(self.explainer, exp_threshold)
            _save_explanation(exp_threshold, path)
            logging.debug(
                "THRESHOLD::Generated; Path %s; SUCCEEDED",
                (path),
            )
            if is_empty_graph(exp_threshold):
                logging.warning(
                    "THRESHOLD::Generated; Path %s; EMPTY GRAPH; FAILED",
                    (path),
                )
                return None
        return exp_threshold

    def get_metric(self, metric: Metric, item: Explanation, path: str):
        if is_exists(path):
            if self.explaining_cfg.explainer.force:
                out_metric = metric.forward(item)
                logging.debug(
                    "METRIC::Generated; Path %s; SUCCEEDED",
                    (path),
                )
            else:
                out_metric = read_json(path)
                logging.debug(
                    "METRIC::Loaded; Path %s; SUCCEEDED",
                    (path),
                )
        else:
            out_metric = metric.forward(item)
            data = {f"{metric.name}": out_metric}
            write_json(data, path)
            if out_metric is None:
                logging.debug(
                    "METRIC::Generated; Path %s; FAILED",
                    (path),
                )
            else:
                logging.debug(
                    "METRIC::Generated; Path %s; SUCCEEDED",
                    (path),
                )
        return out_metric

    def get_stat(self, item: Data, path: str):
        if self.graphstat is None:
            self.load_graphstat()
        if is_exists(path):
            pass
        else:
            if item.num_nodes <= 500:
                stat = self.graphstat(item)
                write_json(stat, path)

    def get_attack(self, attack: Attack, item: Data, path: str):
        if is_exists(path):
            if self.explaining_cfg.explainer.force:
                data_attack = attack.get_attacked_prediction(item)
                logging.debug(
                    "ATTACK::Generated %s; Path %s; SUCCEEDED",
                    (path),
                )

            else:
                data_attack = _load_explanation(path)
                logging.debug(
                    "ATTACK::Generated %s; Path %s; SUCCEEDED",
                    (path),
                )
        else:
            data_attack = attack.get_attacked_prediction(item)
            _save_explanation(data_attack, path)
            logging.debug(
                "ATTACK::Generated %s; Path %s; SUCCEEDED",
                (path),
            )
        return data_attack

    def setup_experiment(self):
        now = datetime.datetime.now()
        self.out_dir = os.path.join(
            self.explaining_cfg.out_dir,
            self.cfg.dataset.name,
            self.model_signature,
        )
        makedirs(self.out_dir)

        now_str = now.strftime("month=%m-day=%d-year=%Y-hour=%H-minute=%M-second=%S")
        set_printing(f"{self.out_dir}/logging-{now_str}.log")

        dump_cfg(self.cfg, os.path.join(self.out_dir, "config.yaml"))
        write_json(self.model_info, os.path.join(self.out_dir, "info.json"))

        self.explainer_path = os.path.join(
            self.out_dir,
            self.explaining_cfg.explainer.name,
            obj_config_to_str(self.explaining_algorithm),
        )
        makedirs(self.explainer_path)
        dump_cfg(
            self.explainer_cfg,
            os.path.join(self.explainer_path, "explainer_cfg.yaml"),
        )
        dump_cfg(
            self.explaining_cfg,
            os.path.join(self.explainer_path, self.explaining_cfg.cfg_dest),
        )

        logging.info("Setting up experiment")
        logging.info("Date and Time: %s", now)
        logging.info("Save experiment to %s", self.out_dir)
        logging.info(
            "####################################################################################"
        )
        logging.info(
            "############################### ORIGINAL CONFIG FILE ###############################"
        )
        logging.info(
            "####################################################################################"
        )
        logging.info(self.cfg)
        logging.info(
            "####################################################################################"
        )
        logging.info(
            "############################### EXPLAINING CONFIG FILE ###############################"
        )
        logging.info(
            "####################################################################################"
        )
        logging.info(self.explaining_cfg)
        logging.info(
            "####################################################################################"
        )
        logging.info(
            "############################### EXPLAINER CONFIG FILE ###############################"
        )
        logging.info(
            "####################################################################################"
        )
        logging.info(self.explainer_cfg)
        logging.info(
            "####################################################################################"
        )
        logging.info(
            "############################### GNN ARCHITECTURE ###############################"
        )
        logging.info(
            "####################################################################################"
        )
        logging.info(self.model)
        logging.info(
            "####################################################################################"
        )
        logging.info(
            "############################### GNN METRICS ###############################"
        )
        logging.info(
            "####################################################################################"
        )
        logging.info(obj_config_to_log(self.model_info))
        logging.info(
            "####################################################################################"
        )
        logging.info(
            "############################### METRICS (EXPLANATION PROCESS)  ###############################"
        )
        logging.info(
            "####################################################################################"
        )
        for metric in self.metrics + self.attacks:
            logging.info(obj_config_to_str(metric))
        for threshold_conf in self.thresholds_configs:
            logging.info(obj_config_to_str(threshold_conf))

        logging.info(
            "####################################################################################"
        )
        logging.info(
            "############################### RUNNING EXPLANATIONS ... ###############################"
        )
        logging.info(
            "####################################################################################"
        )
